"""
This module contains the implementation of the
FR Test Statistic based estimate and the
FNN based estimate for the Bayes Error Rate
"""
from abc import abstractmethod
from typing import Tuple

import numpy as np
import torch
from scipy.sparse import coo_matrix

from daml._internal.metrics.aria.base import _BaseMetric
from daml._internal.metrics.outputs import BEROutput

from .utils import compute_neighbors, get_classes_counts, minimum_spanning_tree


class _MultiClassBer(_BaseMetric):
    def __init__(
        self,
        images: np.ndarray,
        labels: np.ndarray,
        encode: bool = False,
        device: torch.device = torch.device("cpu"),
    ) -> None:
        """Constructor method"""
        super().__init__(images, encode, device=device)
        self.images = images
        self.labels = labels

    @abstractmethod
    def _multiclass_ber(
        self,
        X: np.ndarray,
        y: np.ndarray,
    ) -> Tuple[float, float]:
        """Abstract method for the implementation of multiclass BER calculation"""

    def _evaluate(self) -> BEROutput:
        """
        Return the Bayes Error Rate estimate

        Returns
        -------
        BEROutput
            The estimated upper and lower bounds of the Bayes Error Rate

        Raises
        ------
        ValueError
            If unique classes M < 2
        """
        # Pass X through an autoencoder before evaluating BER
        embeddings = self._encode(self.images)
        ber, ber_lower = self._multiclass_ber(embeddings, self.labels)
        return BEROutput(ber=ber, ber_lower=ber_lower)


class MultiClassBerMST(_MultiClassBer):
    """
    Implements the FR Test Statistic based estimator for the Bayes Error Rate

    Note
    ----
    `Learning to Bound the Multi-class Bayes Error (Th. 3 and Th. 4) <https://arxiv.org/abs/1811.06419>`_
    """  # noqa F401

    def _multiclass_ber(
        self,
        X: np.ndarray,
        y: np.ndarray,
    ) -> Tuple[float, float]:
        """
        Calculates the Bayes Error Rate estimate

        Parameters
        ----------
        X : np.ndarray
            (n_samples x n_features) array of covariates (or image embeddings)
        y : np.ndarray
            n_samples vector of class labels with M unique classes. 2 <= M <= 10

        Returns
        -------
        float
            Estimate of the Bayes Error Rate

        Raises
        ------
        ValueError
            If unique classes M < 2 or M > 10
        """
        M, N = get_classes_counts(y)

        tree = coo_matrix(minimum_spanning_tree(X))
        matches = np.sum([y[tree.row[i]] != y[tree.col[i]] for i in range(N - 1)])
        deltas = matches / (2 * N)
        upper = 2 * deltas
        lower = ((M - 1) / (M)) * (1 - (1 - 2 * ((M) / (M - 1)) * deltas) ** 0.5)
        return upper, lower


class MultiClassBerFNN(_MultiClassBer):
    """
    Implements the KNN Test Statistic based estimator for the Bayes Error Rate

    Parameters
    ----------
    X : np.ndarray
        (n_samples x n_features) array of covariates (or image embeddings)
    y : np.ndarray
        n_samples vector of class labels with M unique classes. 2 <= M <= 10

    Returns
    -------
    float
        Estimate of the Bayes Error Rate

    Raises
    ------
    ValueError
        If unique classes M < 2 or M > 10

    See Also
    --------
    `Learning to Bound the Multi-class Bayes Error (Th. 3 and Th. 4) <https://arxiv.org/abs/1811.06419>`_
    """  # noqa F401

    def _multiclass_ber(
        self,
        X: np.ndarray,
        y: np.ndarray,
    ) -> Tuple[float, float]:
        M, N = get_classes_counts(y)

        # All features belong on second dimension
        X = X.reshape((X.shape[0], -1))
        nn_indices = compute_neighbors(X, X)
        deltas = float(np.count_nonzero(y[nn_indices] - y) / (2 * N))
        upper = 2 * deltas
        lower = ((M - 1) / (M)) * (1 - (1 - 2 * ((M) / (M - 1)) * deltas) ** 0.5)
        return upper, lower



from dataclasses import dataclass
from enum import Enum


class ThresholdType(Enum):
    VALUE = "value"
    PERCENTAGE = "percentage"


@dataclass
class Threshold:
    value: float
    type: ThresholdType



"""
This module contains implementations of Image Outlier Detection methods
created by Alibi Detect
"""

from typing import Optional

import numpy as np

from daml import _alibi_detect
from daml._internal.metrics.alibi_detect.base import (
    AlibiDetectOutlierType,
    _AlibiDetectMetric,
)


class AlibiLLR(_AlibiDetectMetric):
    """
    Log likelihood Ratio (LLR) outlier detector,
    using `alibi-detect llr. <https://docs.seldon.io/projects/alibi-detect/en/latest/examples/od_llr_mnist.html>`_


    The model used by this class is :py:class:`daml.models.LLR`
    """  # noqa E501

    def __init__(self):
        super().__init__(
            alibi_detect_class=_alibi_detect.od.LLR,
            model_param_name="model",
            flatten_dataset=False,
            dataset_type=np.float32,
        )

    def set_prediction_args(
        self,
        outlier_type: Optional[AlibiDetectOutlierType] = None,
        return_instance_score: Optional[bool] = None,
    ) -> None:
        """
        Sets additional arguments to be used during prediction.

        Note
        ----
        Visit `alibi-detect llr <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/llr.html#Detect>`_ for additional information on prediction parameters.
        """  # noqa E501
        self._update_kwargs_with_locals(self._predict_kwargs, **locals())

    @property
    def _default_predict_kwargs(self) -> dict:
        return {
            "outlier_type": AlibiDetectOutlierType.INSTANCE,
            "return_instance_score": True,
            "batch_size": 64,
        }



from functools import partial
from typing import Callable, Type, Union

import numpy as np
import tensorflow as tf

from daml._alibi_detect.utils.prediction import tokenize_transformer


def predict_batch(
    x: Union[list, np.ndarray, tf.Tensor],
    model: Union[Callable, tf.keras.Model],
    batch_size: int = int(1e10),
    preprocess_fn: Callable = None,
    dtype: Union[Type[np.generic], tf.DType] = np.float32,
) -> Union[np.ndarray, tf.Tensor, tuple]:
    """
    Make batch predictions on a model.

    Parameters
    ----------
    x
        Batch of instances.
    model
        tf.keras model or one of the other permitted types defined in Data.
    batch_size
        Batch size used during prediction.
    preprocess_fn
        Optional preprocessing function for each batch.
    dtype
        Model output type, e.g. np.float32 or tf.float32.

    Returns
    -------
    Numpy array, tensorflow tensor or tuples of those with model outputs.
    """
    n = len(x)
    n_minibatch = int(np.ceil(n / batch_size))
    return_np = not isinstance(dtype, tf.DType)
    return_list = False
    preds: Union[list, tuple] = []
    for i in range(n_minibatch):
        istart, istop = i * batch_size, min((i + 1) * batch_size, n)
        x_batch = x[istart:istop]
        if isinstance(preprocess_fn, Callable):  # type: ignore
            x_batch = preprocess_fn(x_batch)
        preds_tmp = model(x_batch)
        if isinstance(preds_tmp, (list, tuple)):
            if len(preds) == 0:  # init tuple with lists to store predictions
                preds = tuple([] for _ in range(len(preds_tmp)))
                return_list = isinstance(preds_tmp, list)
            for j, p in enumerate(preds_tmp):
                preds[j].append(
                    p if not return_np or isinstance(p, np.ndarray) else p.numpy()
                )
        elif isinstance(preds_tmp, (np.ndarray, tf.Tensor)):
            preds.append(
                preds_tmp
                if not return_np or isinstance(preds_tmp, np.ndarray)  # type: ignore
                else preds_tmp.numpy()
            )
        else:
            raise TypeError(
                f"Model output type {type(preds_tmp)} not supported. The model output "
                f"type needs to be one of list, tuple, np.ndarray or tf.Tensor."
            )
    concat = np.concatenate if return_np else tf.concat
    out = (
        tuple(concat(p, axis=0) for p in preds)
        if isinstance(preds, tuple)
        else concat(preds, axis=0)
    )
    if return_list:
        out = list(out)
    return out


def predict_batch_transformer(
    x: Union[list, np.ndarray],
    model: tf.keras.Model,
    tokenizer: Callable,
    max_len: int,
    batch_size: int = int(1e10),
    dtype: Union[Type[np.generic], tf.DType] = np.float32,
) -> Union[np.ndarray, tf.Tensor]:
    """
    Make batch predictions using a transformers tokenizer and model.

    Parameters
    ----------
    x
        Batch of instances.
    model
        Transformer model.
    tokenizer
        Tokenizer for model.
    max_len
        Max token length.
    batch_size
        Batch size.
    dtype
        Model output type, e.g. np.float32 or tf.float32.

    Returns
    -------
    Numpy array or tensorflow tensor with model outputs.
    """
    preprocess_fn = partial(
        tokenize_transformer, tokenizer=tokenizer, max_len=max_len, backend="tf"
    )
    return predict_batch(
        x, model, preprocess_fn=preprocess_fn, batch_size=batch_size, dtype=dtype
    )



"""This module contains dataclasses for each metric output"""

from dataclasses import dataclass
from typing import Iterable

import numpy as np


@dataclass
class OutlierDetectorOutput:
    """
    Dataclass to store outputs from Outlier Detection metrics

    Attributes
    ----------
    is_outlier : List[bool]
        Indicates which elements of the dataset are marked as outliers
    feature_score : np.ndarray
        Feature level outlier scores. Shape = (B, H, W, C)
    instance_score : np.ndarray
        Instance (image) level outlier scores. Shape = (B, )
    """

    is_outlier: Iterable[bool]
    feature_score: np.ndarray
    instance_score: np.ndarray


@dataclass
class DivergenceOutput:
    """
    Dataclass to store outputs from Divergence metrics

    Attributes
    ----------
    dpdivergence : float
        Measure of the distance between two distributions (or datasets).

        For more information about this divergence, its formal definition,
        and its associated estimators
        visit https://arxiv.org/abs/1412.6534.

    error : float
        Number of edges connecting the two datasets
    """

    dpdivergence: float
    error: float

    def __eq__(self, other):
        if isinstance(other, DivergenceOutput):
            if self.dpdivergence == other.dpdivergence and self.error == other.error:
                return True
        return False


@dataclass
class BEROutput:
    """
    Dataclass to store output from BER metrics

    Attributes
    ----------
    ber : float
        The minimum error rate that can be achieved by a classifier
        based on the Bayes decision rule. This is the upper bound.
    ber_lower: float
        Lower bound for the above.
    """

    ber: float
    ber_lower: float


@dataclass
class UAPOutput:
    """
    Dataclass to store output from UAP metrics

    Attributes
    ----------
    uap : float
        The maximum average precision that can be achieved by a detector.
    """

    uap: float



"""
This module contains implementations of Image Outlier Detection methods
created by Alibi Detect
"""

import math
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Optional, Tuple

import numpy as np
import tensorflow as tf

from daml._internal.metrics.outputs import OutlierDetectorOutput
from daml._internal.metrics.types import Threshold, ThresholdType
from daml._internal.models.tensorflow.alibi import create_model


class AlibiDetectOutlierType(str, Enum):
    INSTANCE = "instance"
    FEATURE = "feature"

    def __str__(self) -> str:
        return str.__str__(self)  # pragma: no cover


class _AlibiDetectMetric(ABC):
    """
    Base class for all outlier detection metrics in alibi-detect

    Attributes
    ----------
    detector : Any, default None
        A model used for outlier detection after being trained on clean data
    """

    @staticmethod
    def _update_kwargs_with_locals(kwargs_to_update, **kwargs):
        kwargs_to_update.update(
            {k: v for k, v in kwargs.items() if k != "self" and v is not None}
        )

    @abstractmethod
    def set_prediction_args(self) -> None:
        """Abstract method to set specific prediction arguments for the detector"""

    @property
    @abstractmethod
    def _default_predict_kwargs(self) -> dict:
        """Abstract method for the default prediction arguments for the detector"""

    def __init__(
        self,
        alibi_detect_class: type,
        model_param_name: str,
        flatten_dataset: bool,
        dataset_type: Optional[type] = None,
    ):
        """
        Constructor method

        Parameters
        ----------
        alibi_detect_class : type
            This is the alibi_detect outlier detection class to instantiate
        flatten_dataset : bool, default False
            Flag to flatten a dataset to shape (B, H * W * C)
        dataset_type : Optional[type], default None
            Type used for formatting the dataset
        """
        self.detector: Any = None
        self.is_trained: bool = False

        self._alibi_detect_class = alibi_detect_class
        self._model_param_name = model_param_name
        self._flatten_dataset = flatten_dataset
        self._dataset_type = dataset_type

        self._model_kwargs = dict()
        self._predict_kwargs = self._default_predict_kwargs
        self._default_batch_size = 64

        self._input_shape: Tuple[int, int, int]

    # Train the alibi-detect metric on dataset
    def fit_dataset(
        self,
        images: np.ndarray,
        epochs: int = 3,
        threshold: Threshold = Threshold(95.0, ThresholdType.PERCENTAGE),
        batch_size: Optional[int] = None,
        verbose: bool = False,
    ) -> None:
        """
        Trains a model on a dataset containing that can be used
        for the detection of outliers in :meth:`evaluate`

        Parameters
        ----------
        images : np.ndarray
            A numpy array of images for the model to train on
        epochs : int, default 3
            Number of epochs to train the detector for
        threshold : Threshold, default Threshold(95.0, ThresholdType.PERCENTAGE)
            Sets the expected threshold of an outlier in the dataset
        batch_size : Optional[int], default None
            Batch size override to use during training
        verbose : bool, default False
            Flag to output logs from Alibi-Detect verbose mode

        Raises
        ------
        TypeError
            If the detector has not been initialized or loaded from path

        Note
        ----
        The supplied dataset should contain no outliers for maximum benefit
        """
        tf.keras.backend.clear_session()
        if self.detector is None:
            self._input_shape = images[0].shape
            model = create_model(type(self).__name__, self._input_shape)
            self.detector = self._alibi_detect_class(
                threshold=0,
                **{self._model_param_name: model},
            )

        # Autoencoders only need images, so extract from dataset and format
        images = self._format_images(images)
        # Train the autoencoder using only the formatted images
        self.detector.fit(
            images,
            epochs=epochs,
            batch_size=batch_size or self._default_batch_size,
            verbose=verbose,
        )

        # if save_path: save_detector(self.detector, save_path)
        self.is_trained: bool = True

        if threshold.type is ThresholdType.PERCENTAGE:
            self.detector.infer_threshold(
                images,
                threshold_perc=threshold.value,
                batch_size=batch_size or self._default_batch_size,
            )
        else:
            self.detector.threshold = threshold.value

    def export_model(self, path: str):
        if self.detector is None:
            raise RuntimeError("The model must be initialized first")
        if self._model_param_name not in self.detector.__dict__.keys():
            raise ValueError("Member not found in detector")
        getattr(self.detector, self._model_param_name).save(path)

    def _check_dtype(self, images: np.ndarray):
        """
        Check if the dataset dtype fits with the required dtype of the model.
        None is used for any accepted type. Can extend to check for multiple types.

        Parameters
        ----------
        images : np.ndarray
            A numpy ndarray of images in the shape (B, H, W, C)
        dtype : Type
            The preferred dtype of the array

        Raises
        ------
        TypeError
            If the dataset is not of type np.ndarray

            If the dataset is not of the preferred dtype
        """

        if self._dataset_type is None:
            return

        # Use ndarray type conversion function
        if not isinstance(images, np.ndarray):
            raise TypeError("Dataset should be of type: np.ndarray")

        if not images.dtype.type == self._dataset_type:
            raise TypeError(
                f"Dataset values should be of type {self._dataset_type}, \
                not {images.dtype.type}"
            )

    def _check_image_shape(self, images: np.ndarray):
        """
        Verifies that a dataset has the same shape as a reference shape.
        Use this when fitting or evaluating a model on a dataset, to
        ensure that the dataset's dims are compatible with that which the model
        was built for.

        Parameters
        ----------
        image : np.ndarray
            An array of images in shape (B, H, W, C)

        Raises
        -------
        TypeError
            Raised if the dims of the last 3 channels of dataset
            don't line up with reference_input_shape
        """
        if self.detector is None:
            return
        if images.shape[-3:] != self._input_shape:
            raise ValueError(
                f"Model was initialized on dataset shape \
                (W,H,C)={self._input_shape}, \
                but provided dataset has shape (W,H,C)={images.shape[-3:]}"
            )

    def _format_images(self, images: np.ndarray) -> np.ndarray:
        """
        Formats images such that it fits the required datatype and shape

        Parameters
        ----------
        images : np.ndarray
            A numpy array of images in shape (B, H, W, C)

        Returns
        -------
        np.ndarray
            A numpy array of formatted images

        Note
        ----
        - Some metric libraries want a dataset in a particular format
            (e.g. float32, or flattened)
        - Override this to set the standard dataset formatting.
        """
        self._check_image_shape(images)
        self._check_dtype(images)

        if self._flatten_dataset:
            images = np.reshape(images, (len(images), math.prod(np.shape(images[0]))))
        return images

    def evaluate(
        self,
        images: np.ndarray,
    ) -> OutlierDetectorOutput:
        """
        Evaluate the outlier detector metric on a dataset.

        Parameters
        ----------
        images : np.ndarray
            The images to detect outliers on.

        Returns
        -------
        OutlierDetectorOutput
            Outlier mask, and associated feature and instance scores

        """
        if self.detector is None or not self.is_trained:
            raise TypeError(
                "Error: tried to evaluate a metric that is not trained. \
                    Try calling metric.fit_dataset(data)"
            )

        # Cast and flatten images
        images = self._format_images(images)

        self._predict_kwargs.update({"X": images})

        predictions = self.detector.predict(**self._predict_kwargs)

        output = OutlierDetectorOutput(
            is_outlier=predictions["data"]["is_outlier"].tolist(),
            feature_score=predictions["data"]["feature_score"],
            instance_score=predictions["data"]["instance_score"],
        )

        return output



from typing import Optional

import numpy as np

from daml import _alibi_detect
from daml._internal.metrics.alibi_detect.base import _AlibiDetectMetric


class AlibiVAEGMM(_AlibiDetectMetric):
    """
    Variational Gaussian Mixture Model Autoencoder-based outlier detector,
    using `alibi-detect vaegmm. <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/vaegmm.html>`_


    The model used by this class is :py:class:`daml.models.VAEGMM`
    """  # noqa E501

    def __init__(self):
        super().__init__(
            alibi_detect_class=_alibi_detect.od.OutlierVAEGMM,
            model_param_name="vaegmm",
            flatten_dataset=True,
            dataset_type=np.float32,
        )

    def set_prediction_args(
        self,
        return_instance_score: Optional[bool] = None,
    ) -> None:
        """
        Sets additional arguments to be used during prediction.

        Note
        ----
        Visit `alibi-detect vaegmm <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/vaegmm.html#Detect>`_ for additional information on prediction parameters.
        """  # noqa E501
        self._update_kwargs_with_locals(self._predict_kwargs, **locals())

    @property
    def _default_predict_kwargs(self) -> dict:
        return {"return_instance_score": True, "batch_size": 64}



"""
This module contains the implementation of the
FR Test Statistic based estimate for the upperbound
average precision (UAP_MST) and empirical mean
precision (UAP_EMP)
"""

import numpy as np
from sklearn.metrics import average_precision_score

from daml._internal.metrics.aria.base import _BaseMetric
from daml._internal.metrics.outputs import UAPOutput

from .utils import get_classes_counts, minimum_spanning_tree


class UAP_MST(_BaseMetric):
    def __init__(self, images: np.ndarray, labels: np.ndarray) -> None:
        """
        Parameters
        ----------
        images : np.ndarray
            A numpy array containing (n_samples x n_features) array of (padded) instance
            embeddings.
        labels : np.ndarray
            A numpy array containing n_samples vector of class labels with M unique
            classes.
        """
        super().__init__(images, False)
        self.labels = labels

    def _evaluate(self) -> UAPOutput:
        """
        Upperbound Average Precision estimate

        Returns
        -------
        UAPOutput
            The estimated UAP
        """
        uap = self._uap(self.data, self.labels)
        return UAPOutput(uap=uap)

    def _uap(
        self,
        X: np.ndarray,
        y: np.ndarray,
    ) -> float:
        """
        Calculates the UAP estimate

        Parameters
        ----------
        X : np.ndarray
            (n_samples x n_features) array of (padded) instance embeddings
        y : np.ndarray
            n_samples vector of class labels with M unique classes. 2 <= M <= 10

        Returns
        -------
        float
            Estimate of the UAP

        Raises
        ------
        ValueError
            If unique classes M < 2 or M > 10
        """
        M, N = get_classes_counts(y)

        tree = minimum_spanning_tree(X).todense()
        conf_mat = np.zeros((M, M))
        for i in range(len(tree)):
            edges = np.where(tree[i, :] != 0)[1]
            for j in range(len(edges)):
                conf_mat[int(y[i]), int(y[edges[j]])] = (
                    conf_mat[int(y[i]), int(y[edges[j]])] + 1
                )
        probs = np.zeros((M, M))
        for i in range(M):
            probs[i, :] = conf_mat[i, :] / np.sum(conf_mat[i, :])
        lab_inds = np.zeros((N, M))
        prob_inds = np.zeros((N, M))
        for i in range(N):
            lab_inds[i, int(y[i])] = 1
            prob_inds[i, :] = probs[int(y[i]), :]
        uap = average_precision_score(lab_inds, prob_inds)
        return float(uap)


class UAP_EMP(_BaseMetric):
    def __init__(
        self,
        labels: np.ndarray,
        scores: np.ndarray,
    ) -> None:
        """
        Parameters
        ----------
        labels : np.ndarray
            A numpy array of n_samples of class labels with M unique classes.

        scores : np.ndarray
            A 2D array of class probabilities per image
        """
        super().__init__(np.ndarray([]), encode=False)
        self.labels = labels
        self.scores = scores

    def _evaluate(self) -> UAPOutput:
        """
        Returns
        -------
        UAPOutput
            The estimated UAP
        """
        uap = float(
            average_precision_score(self.labels, self.scores, average="weighted")
        )
        return UAPOutput(uap=uap)



import logging
from typing import Callable, Dict, Tuple

import numpy as np
import tensorflow as tf

from daml._alibi_detect.base import (
    BaseDetector,
    FitMixin,
    ThresholdMixin,
    outlier_prediction_dict,
)
from daml._alibi_detect.models.tensorflow.autoencoder import AEGMM, eucl_cosim_features
from daml._alibi_detect.models.tensorflow.gmm import gmm_energy, gmm_params
from daml._alibi_detect.models.tensorflow.losses import loss_aegmm
from daml._alibi_detect.models.tensorflow.trainer import trainer
from daml._alibi_detect.utils._types import OptimizerTF
from daml._alibi_detect.utils.tensorflow.prediction import predict_batch

logger = logging.getLogger(__name__)


class OutlierAEGMM(BaseDetector, FitMixin, ThresholdMixin):
    def __init__(
        self,
        threshold: float = None,
        aegmm: tf.keras.Model = None,
        encoder_net: tf.keras.Model = None,
        decoder_net: tf.keras.Model = None,
        gmm_density_net: tf.keras.Model = None,
        n_gmm: int = None,
        recon_features: Callable = eucl_cosim_features,
        data_type: str = None,
    ) -> None:
        """
        AEGMM-based outlier detector.

        Parameters
        ----------
        threshold
            Threshold used for outlier score to determine outliers.
        aegmm
            A trained tf.keras model if available.
        encoder_net
            Layers for the encoder wrapped in a tf.keras.Sequential class if no 'aegmm' is specified.
        decoder_net
            Layers for the decoder wrapped in a tf.keras.Sequential class if no 'aegmm' is specified.
        gmm_density_net
            Layers for the GMM network wrapped in a tf.keras.Sequential class.
        n_gmm
            Number of components in GMM.
        recon_features
            Function to extract features from the reconstructed instance by the decoder.
        data_type
            Optionally specify the data type (tabular, image or time-series). Added to metadata.
        """
        super().__init__()

        if threshold is None:
            logger.warning(
                "No threshold level set. Need to infer threshold using `infer_threshold`."
            )

        self.threshold = threshold

        # check if model can be loaded, otherwise initialize AEGMM model
        if isinstance(aegmm, tf.keras.Model):
            self.aegmm = aegmm
        elif (
            isinstance(encoder_net, tf.keras.Sequential)
            and isinstance(decoder_net, tf.keras.Sequential)
            and isinstance(gmm_density_net, tf.keras.Sequential)
        ):
            self.aegmm = AEGMM(
                encoder_net, decoder_net, gmm_density_net, n_gmm, recon_features
            )
        else:
            raise TypeError(
                "No valid format detected for `aegmm` (tf.keras.Model) "
                "or `encoder_net`, `decoder_net` and `gmm_density_net` (tf.keras.Sequential)."
            )

        # set metadata
        self.meta["detector_type"] = "outlier"
        self.meta["data_type"] = data_type
        self.meta["online"] = False

        self.phi, self.mu, self.cov, self.L, self.log_det_cov = (
            None,
            None,
            None,
            None,
            None,
        )

    def fit(
        self,
        X: np.ndarray,
        loss_fn: tf.keras.losses = loss_aegmm,
        w_energy: float = 0.1,
        w_cov_diag: float = 0.005,
        optimizer: OptimizerTF = tf.keras.optimizers.Adam,
        epochs: int = 20,
        batch_size: int = 64,
        verbose: bool = True,
        log_metric: Tuple[str, "tf.keras.metrics"] = None,
        callbacks: tf.keras.callbacks = None,
    ) -> None:
        """
        Train AEGMM model.

        Parameters
        ----------
        X
            Training batch.
        loss_fn
            Loss function used for training.
        w_energy
            Weight on sample energy loss term if default `loss_aegmm` loss fn is used.
        w_cov_diag
            Weight on covariance regularizing loss term if default `loss_aegmm` loss fn is used.
        optimizer
            Optimizer used for training.
        epochs
            Number of training epochs.
        batch_size
            Batch size used for training.
        verbose
            Whether to print training progress.
        log_metric
            Additional metrics whose progress will be displayed if verbose equals True.
        callbacks
            Callbacks used during training.
        """
        # train arguments
        args = [self.aegmm, loss_fn, X]
        optimizer = optimizer() if isinstance(optimizer, type) else optimizer
        kwargs = {
            "optimizer": optimizer,
            "epochs": epochs,
            "batch_size": batch_size,
            "verbose": verbose,
            "log_metric": log_metric,
            "callbacks": callbacks,
            "loss_fn_kwargs": {"w_energy": w_energy, "w_cov_diag": w_cov_diag},
        }

        # train
        trainer(*args, **kwargs)

        # set GMM parameters
        x_recon, z, gamma = self.aegmm(X)
        self.phi, self.mu, self.cov, self.L, self.log_det_cov = gmm_params(z, gamma)

    def infer_threshold(
        self, X: np.ndarray, threshold_perc: float = 95.0, batch_size: int = int(1e10)
    ) -> None:
        """
        Update threshold by a value inferred from the percentage of instances considered to be
        outliers in a sample of the dataset.

        Parameters
        ----------
        X
            Batch of instances.
        threshold_perc
            Percentage of X considered to be normal based on the outlier score.
        batch_size
            Batch size used when making predictions with the AEGMM.
        """
        # compute outlier scores
        iscore = self.score(X, batch_size=batch_size)

        # update threshold
        self.threshold = np.percentile(iscore, threshold_perc)

    def score(self, X: np.ndarray, batch_size: int = int(1e10)) -> np.ndarray:
        """
        Compute outlier scores.

        Parameters
        ----------
        X
            Batch of instances to analyze.
        batch_size
            Batch size used when making predictions with the AEGMM.

        Returns
        -------
        Array with outlier scores for each instance in the batch.
        """
        _, z, _ = predict_batch(X, self.aegmm, batch_size=batch_size)
        energy, _ = gmm_energy(
            z, self.phi, self.mu, self.cov, self.L, self.log_det_cov, return_mean=False
        )
        return energy.numpy()

    def predict(
        self,
        X: np.ndarray,
        batch_size: int = int(1e10),
        return_instance_score: bool = True,
    ) -> Dict[Dict[str, str], Dict[np.ndarray, np.ndarray]]:
        """
        Compute outlier scores and transform into outlier predictions.

        Parameters
        ----------
        X
            Batch of instances.
        batch_size
            Batch size used when making predictions with the AEGMM.
        return_instance_score
            Whether to return instance level outlier scores.

        Returns
        -------
        Dictionary containing ``'meta'`` and ``'data'`` dictionaries.
            - ``'meta'`` has the model's metadata.
            - ``'data'`` contains the outlier predictions and instance level outlier scores.
        """
        # compute outlier scores
        iscore = self.score(X, batch_size=batch_size)

        # values above threshold are outliers
        outlier_pred = (iscore > self.threshold).astype(int)

        # populate output dict
        od = outlier_prediction_dict()
        od["meta"] = self.meta
        od["data"]["is_outlier"] = outlier_pred
        if return_instance_score:
            od["data"]["instance_score"] = iscore
        return od



from typing import Optional, Tuple, Union

import numpy as np
from torch.utils.data import Dataset


class DamlDataset(Dataset):
    """Holds the arrays of images and labels"""

    def __init__(
        self,
        images: np.ndarray,
        labels: Optional[np.ndarray] = None,
        boxes: Optional[np.ndarray] = None,
    ) -> None:
        self._images: np.ndarray = images
        self._labels = labels
        self._boxes = boxes

        self._validate()

    def __len__(self) -> int:
        return len(self._images)

    def __getitem__(self, index) -> Union[np.ndarray, Tuple]:
        image: np.ndarray = self._images[index]

        # Return image if no other attributes
        if self._labels is None:
            return image

        labels: np.ndarray = self._labels[index]
        # Return image and label for image classification
        if self._boxes is None:
            return image, labels

        # Return images, labels, boxes for object detection
        boxes: np.ndarray = self.boxes[index]
        return image, labels, boxes

    @property
    def images(self) -> np.ndarray:
        return self._images

    @property
    def labels(self) -> np.ndarray:
        if self._labels is None:
            return np.array([])
        else:
            return self._labels

    @property
    def boxes(self) -> np.ndarray:
        if self._boxes is None:
            return np.array([])
        else:
            return self._boxes

    def _validate(self):
        if self._labels is not None:
            if self._boxes is None:
                assert len(self._images) == len(self._labels)
            else:
                assert len(self._boxes) == len(self._labels)
        else:
            assert self._boxes is None



from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar, Union

import numpy as np
import torch
import torch.nn as nn

from daml._internal.metrics.aria.utils import permute_to_torch, pytorch_to_numpy
from daml._internal.models.pytorch.autoencoder import AERunner, AETrainer

TOutput = TypeVar("TOutput")


class _BaseMetric(ABC, Generic[TOutput]):
    """Abstract base class for metrics"""

    def __init__(
        self,
        data: np.ndarray,
        encode: bool,
        model: Optional[nn.Module] = None,
        fit: Optional[bool] = None,
        epochs: Optional[int] = None,
        device: torch.device = torch.device("cpu"),
    ):
        self.data = data
        self.encode = encode
        self.model = model

        # TODO: Model training args will move out of metrics
        self.fit = fit
        self.epochs = epochs
        self._device = device

        self._validate_args()

    def evaluate(self) -> TOutput:
        # TODO: Split model training from metrics and make evaluate the abstractmethod
        if self.fit or self.encode:
            self._fit()
        return self._evaluate()

    @abstractmethod
    def _evaluate(self) -> TOutput:
        """Abstract method to calculate metric based off of constructor parameters"""

    def _fit(self) -> None:
        """
        Trains a model on a dataset to be used during calculation of metrics.
        """
        self._validate_args()
        if isinstance(self.model, nn.Module) and not self.fit:
            self.model = AERunner(model=self.model, device=self._device)
        else:
            images = permute_to_torch(self.data)
            if isinstance(self.model, nn.Module):
                self.model = AETrainer(model=self.model, device=self._device)
            else:
                self.model = AETrainer(channels=images.shape[1], device=self._device)

    def _encode(self, images: Union[torch.Tensor, np.ndarray]) -> np.ndarray:
        """
        Takes an array or tensor of images and returns a tensor that is encoded if
        `self.encode` is True
        """
        self._validate_args()

        if self.encode:
            if not isinstance(images, torch.Tensor):
                images = permute_to_torch(images)
            assert isinstance(self.model, nn.Module)
            images = self.model.encode(images).detach().cpu().numpy()

        return images if isinstance(images, np.ndarray) else pytorch_to_numpy(images)

    def _validate_args(self):
        if self.model is not None and not isinstance(self.model, nn.Module):
            raise TypeError(
                f"Given model is of type {type(self.model)}, expected nn.Module"
            )

        if not (self.model is None and self.fit is None and self.epochs is None):
            if self.model is None:
                raise ValueError(
                    "Must specify `model` if model arguments are provided."
                )
            if self.fit is None:
                raise ValueError("Must specify `fit` if model arguments are provided.")
            if self.epochs is None and self.fit:
                raise ValueError("Must specify `epochs` to fit model.")

        if not self.encode and self.model is not None:
            raise ValueError("Model not used if `encode` is `false`.")



import math
from typing import Tuple

from tensorflow.keras import Sequential
from tensorflow.keras.layers import (
    Conv2D,
    Conv2DTranspose,
    Dense,
    Flatten,
    InputLayer,
    Reshape,
)
from tensorflow.math import reduce_prod
from tensorflow.nn import relu, softmax, tanh

from daml._alibi_detect.models.tensorflow.autoencoder import AE, AEGMM, VAE, VAEGMM
from daml._alibi_detect.models.tensorflow.pixelcnn import PixelCNN


def _get_default_encoder_net(input_shape: Tuple[int, int, int], encoding_dim: int):
    return Sequential(
        [
            InputLayer(input_shape=input_shape),
            Conv2D(64, 4, strides=2, padding="same", activation=relu),
            Conv2D(128, 4, strides=2, padding="same", activation=relu),
            Conv2D(512, 4, strides=2, padding="same", activation=relu),
            Flatten(),
            Dense(encoding_dim),
        ]
    )


def _get_default_decoder_net(input_shape: Tuple[int, int, int], encoding_dim: int):
    return Sequential(
        [
            InputLayer(input_shape=(encoding_dim,)),
            Dense(4 * 4 * 128),
            Reshape(target_shape=(4, 4, 128)),
            Conv2DTranspose(256, 4, strides=2, padding="same", activation=relu),
            Conv2DTranspose(64, 4, strides=2, padding="same", activation=relu),
            Flatten(),
            Dense(math.prod(input_shape)),
            Reshape(target_shape=input_shape),
        ]
    )


def create_model(
    class_name: str,
    input_shape: Tuple[int, int, int],
    encoding_dim: int = 1024,
    n_gmm: int = 2,
    aegmm_latent_dim: int = 1,
    vaegmm_latent_dim: int = 2,
):
    if class_name == "AlibiAE":
        return AE(
            _get_default_encoder_net(input_shape, encoding_dim),
            _get_default_decoder_net(input_shape, encoding_dim),
        )

    if class_name == "AlibiVAE":
        return VAE(
            _get_default_encoder_net(input_shape, encoding_dim),
            _get_default_decoder_net(input_shape, encoding_dim),
            encoding_dim,
        )

    if class_name == "AlibiAEGMM":
        n_features = reduce_prod(input_shape)
        n_gmm = 2  # nb of components in GMM
        # The outlier detector is an encoder/decoder architecture
        # Here we define the encoder
        encoder_net = Sequential(
            [
                InputLayer(input_shape=(n_features,)),
                Dense(60, activation=tanh),
                Dense(30, activation=tanh),
                Dense(10, activation=tanh),
                Dense(aegmm_latent_dim, activation=None),
            ]
        )
        # Here we define the decoder
        decoder_net = Sequential(
            [
                InputLayer(input_shape=(aegmm_latent_dim,)),
                Dense(10, activation=tanh),
                Dense(30, activation=tanh),
                Dense(60, activation=tanh),
                Dense(n_features, activation=None),
            ]
        )
        # GMM autoencoders have a density network too
        gmm_density_net = Sequential(
            [
                InputLayer(input_shape=(aegmm_latent_dim + 2,)),
                Dense(10, activation=tanh),
                Dense(n_gmm, activation=softmax),
            ]
        )
        return AEGMM(
            encoder_net=encoder_net,
            decoder_net=decoder_net,
            gmm_density_net=gmm_density_net,
            n_gmm=n_gmm,
        )

    if class_name == "AlibiVAEGMM":
        n_features = reduce_prod(input_shape)

        # The outlier detector is an encoder/decoder architecture
        # Here we define the encoder
        encoder_net = Sequential(
            [
                InputLayer(input_shape=(n_features,)),
                Dense(20, activation=relu),
                Dense(15, activation=relu),
                Dense(7, activation=relu),
            ]
        )
        # Here we define the decoder
        decoder_net = Sequential(
            [
                InputLayer(input_shape=(vaegmm_latent_dim,)),
                Dense(7, activation=relu),
                Dense(15, activation=relu),
                Dense(20, activation=relu),
                Dense(n_features, activation=None),
            ]
        )
        # GMM autoencoders have a density network too
        gmm_density_net = Sequential(
            [
                InputLayer(input_shape=(vaegmm_latent_dim + 2,)),
                Dense(10, activation=relu),
                Dense(n_gmm, activation=softmax),
            ]
        )
        return VAEGMM(
            encoder_net=encoder_net,
            decoder_net=decoder_net,
            gmm_density_net=gmm_density_net,
            n_gmm=n_gmm,
            latent_dim=vaegmm_latent_dim,
        )

    if class_name == "AlibiLLR":
        return PixelCNN(
            image_shape=input_shape,
            num_resnet=5,
            num_hierarchies=2,
            num_filters=32,
            num_logistic_mix=1,
            receptive_field_dims=(3, 3),
            dropout_p=0.3,
            l2_weight=0.0,
        )

    raise TypeError("Unknown alibi_detect class specified: ", class_name)



from typing import Optional, Union

import torch
import torch.nn as nn
from torch.optim import Adam
from torch.utils.data import DataLoader

torch.manual_seed(0)


class AERunner:
    def __init__(self, model: nn.Module, device: Union[str, torch.device] = "cpu"):
        self._model = model
        self._device = device

        self._model.to(self._device)

    def encode(self, x) -> torch.Tensor:
        with torch.no_grad():
            if getattr(self._model, "encode", None) is not None:
                return self._model.encode(x)
            else:  # Call model.forward() if no encode function
                return self._model(x)

    def __call__(self, x):
        with torch.no_grad():
            x = self._model(x)
        return x


class AETrainer(AERunner):
    def __init__(
        self,
        model: Optional[nn.Module] = None,
        channels: Optional[int] = None,
        device: Union[str, torch.device] = "cpu",
    ):
        # Channel check only if no model given
        if model is None:
            if channels is None:
                raise TypeError("Channels must be set for default model")
            _model = AriaAutoencoder(channels)
        else:
            _model = model
        super().__init__(_model, device=device)

    def train(self, dataset, epochs: int = 100, batch_size=8):
        dl = DataLoader(dataset, batch_size=batch_size)

        opt = Adam(self._model.parameters(), lr=0.001)
        criterion = nn.MSELoss().to(self._device)
        loss = 0
        for _ in range(epochs):
            for batch in dl:
                imgs = batch[0].to(self._device)  # (imgs, labels, bboxes)
                opt.zero_grad()
                pred = self._model(imgs)
                loss = criterion(pred, imgs)
                loss.backward()
                opt.step()


class AriaAutoencoder(nn.Module):
    def __init__(self, channels=3):
        super().__init__()
        self.encoder = Encoder(channels)
        self.decoder = Decoder(channels)

    def forward(self, x):
        x = self.encoder(x)
        x = self.decoder(x)
        return x

    def encode(self, x):
        return self.encoder(x)


class Encoder(nn.Module):
    def __init__(self, channels=3):
        super().__init__()
        self.encoder = nn.Sequential(
            nn.Conv2d(channels, 256, 2, stride=1, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(256, 128, 2, stride=1, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(128, 64, 2, stride=1),
        )

    def forward(self, x):
        return self.encoder(x)


class Decoder(nn.Module):
    def __init__(self, channels):
        super().__init__()
        self.decoder = nn.Sequential(
            nn.ConvTranspose2d(64, 128, 2, stride=1),
            nn.ReLU(),
            nn.ConvTranspose2d(128, 256, 2, stride=2),
            nn.ReLU(),
            nn.ConvTranspose2d(256, channels, 2, stride=2),
            nn.Sigmoid(),
        )

    def forward(self, x):
        return self.decoder(x)



from typing import Callable, List, Tuple

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import (
    LSTM,
    Bidirectional,
    Concatenate,
    Dense,
    Flatten,
    Layer,
)

from daml._alibi_detect.utils.tensorflow.distance import relative_euclidean_distance


class Sampling(Layer):
    """Reparametrization trick. Uses (z_mean, z_log_var) to sample the latent vector z."""

    def call(self, inputs: Tuple[tf.Tensor, tf.Tensor]) -> tf.Tensor:
        """
        Sample z.

        Parameters
        ----------
        inputs
            Tuple with mean and log variance.

        Returns
        -------
        Sampled vector z.
        """
        z_mean, z_log_var = inputs
        batch = tf.shape(z_mean)[0]
        dim = tf.shape(z_mean)[1]
        epsilon = tf.keras.backend.random_normal(shape=(batch, dim))
        return z_mean + tf.exp(0.5 * z_log_var) * epsilon


class EncoderVAE(Layer):
    def __init__(
        self, encoder_net: tf.keras.Model, latent_dim: int, name: str = "encoder_vae"
    ) -> None:
        """
        Encoder of VAE.

        Parameters
        ----------
        encoder_net
            Layers for the encoder wrapped in a tf.keras.Sequential class.
        latent_dim
            Dimensionality of the latent space.
        name
            Name of encoder.
        """
        super(EncoderVAE, self).__init__(name=name)
        self.encoder_net = encoder_net
        self.fc_mean = Dense(latent_dim, activation=None)
        self.fc_log_var = Dense(latent_dim, activation=None)
        self.sampling = Sampling()

    def call(self, x: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        x = self.encoder_net(x)
        if len(x.shape) > 2:
            x = Flatten()(x)
        z_mean = self.fc_mean(x)
        z_log_var = self.fc_log_var(x)
        z = self.sampling((z_mean, z_log_var))
        return z_mean, z_log_var, z


class Decoder(Layer):
    def __init__(self, decoder_net: tf.keras.Model, name: str = "decoder") -> None:
        """
        Decoder of (V)AE.

        Parameters
        ----------
        decoder_net
            Layers for the decoder wrapped in a tf.keras.Sequential class.
        name
            Name of decoder.
        """
        super(Decoder, self).__init__(name=name)
        self.decoder_net = decoder_net

    def call(self, x: tf.Tensor) -> tf.Tensor:
        return self.decoder_net(x)


class VAE(tf.keras.Model):
    def __init__(
        self,
        encoder_net: tf.keras.Model,
        decoder_net: tf.keras.Model,
        latent_dim: int,
        beta: float = 1.0,
        name: str = "vae",
    ) -> None:
        """
        Combine encoder and decoder in VAE.

        Parameters
        ----------
        encoder_net
            Layers for the encoder wrapped in a tf.keras.Sequential class.
        decoder_net
            Layers for the decoder wrapped in a tf.keras.Sequential class.
        latent_dim
            Dimensionality of the latent space.
        beta
            Beta parameter for KL-divergence loss term.
        name
            Name of VAE model.
        """
        super(VAE, self).__init__(name=name)
        self.encoder = EncoderVAE(encoder_net, latent_dim)
        self.decoder = Decoder(decoder_net)
        self.beta = beta
        self.latent_dim = latent_dim

    def call(self, x: tf.Tensor) -> tf.Tensor:
        z_mean, z_log_var, z = self.encoder(x)
        x_recon = self.decoder(z)
        # add KL divergence loss term
        kl_loss = -0.5 * tf.reduce_mean(
            z_log_var - tf.square(z_mean) - tf.exp(z_log_var) + 1
        )
        self.add_loss(self.beta * kl_loss)
        return x_recon


class EncoderAE(Layer):
    def __init__(self, encoder_net: tf.keras.Model, name: str = "encoder_ae") -> None:
        """
        Encoder of AE.

        Parameters
        ----------
        encoder_net
            Layers for the encoder wrapped in a tf.keras.Sequential class.
        name
            Name of encoder.
        """
        super(EncoderAE, self).__init__(name=name)
        self.encoder_net = encoder_net

    def call(self, x: tf.Tensor) -> tf.Tensor:
        return self.encoder_net(x)


class AE(tf.keras.Model):
    def __init__(
        self, encoder_net: tf.keras.Model, decoder_net: tf.keras.Model, name: str = "ae"
    ) -> None:
        """
        Combine encoder and decoder in AE.

        Parameters
        ----------
        encoder_net
            Layers for the encoder wrapped in a tf.keras.Sequential class.
        decoder_net
            Layers for the decoder wrapped in a tf.keras.Sequential class.
        name
            Name of autoencoder model.
        """
        super(AE, self).__init__(name=name)
        self.encoder = EncoderAE(encoder_net)
        self.decoder = Decoder(decoder_net)

    def call(self, x: tf.Tensor) -> tf.Tensor:
        z = self.encoder(x)
        x_recon = self.decoder(z)
        return x_recon


class EncoderLSTM(Layer):
    def __init__(self, latent_dim: int, name: str = "encoder_lstm") -> None:
        """
        Bidirectional LSTM encoder.

        Parameters
        ----------
        latent_dim
            Latent dimension. Must be an even number given the bidirectional encoder.
        name
            Name of encoder.
        """
        super(EncoderLSTM, self).__init__(name=name)
        self.encoder_net = Bidirectional(
            LSTM(latent_dim // 2, return_state=True, return_sequences=True)
        )

    def call(self, x: tf.Tensor) -> Tuple[tf.Tensor, List[tf.Tensor]]:
        enc_out, fwd_h, fwd_c, bwd_h, bwd_c = self.encoder_net(x)
        h = Concatenate()([fwd_h, bwd_h])
        c = Concatenate()([fwd_c, bwd_c])
        return enc_out, [h, c]


class DecoderLSTM(Layer):
    def __init__(
        self,
        latent_dim: int,
        output_dim: int,
        output_activation: str = None,
        name: str = "decoder_lstm",
    ) -> None:
        """
        LSTM decoder.

        Parameters
        ----------
        latent_dim
            Latent dimension.
        output_dim
            Decoder output dimension.
        output_activation
            Activation used in the Dense output layer.
        name
            Name of decoder.
        """
        super(DecoderLSTM, self).__init__(name=name)
        self.decoder_net = LSTM(latent_dim, return_state=True, return_sequences=True)
        self.dense = Dense(output_dim, activation=output_activation)

    def call(
        self, x: tf.Tensor, init_state: List[tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor, List[tf.Tensor]]:
        x, h, c = self.decoder_net(x, initial_state=init_state)
        dec_out = self.dense(x)
        return dec_out, x, [h, c]


class Seq2Seq(tf.keras.Model):
    def __init__(
        self,
        encoder_net: EncoderLSTM,
        decoder_net: DecoderLSTM,
        threshold_net: tf.keras.Model,
        n_features: int,
        score_fn: Callable = tf.math.squared_difference,
        beta: float = 1.0,
        name: str = "seq2seq",
    ) -> None:
        """
        Sequence-to-sequence model.

        Parameters
        ----------
        encoder_net
            Encoder network.
        decoder_net
            Decoder network.
        threshold_net
            Regression network used to estimate threshold.
        n_features
            Number of features.
        score_fn
            Function used for outlier score.
        beta
            Weight on the threshold estimation loss term.
        name
            Name of the seq2seq model.
        """
        super(Seq2Seq, self).__init__(name=name)
        self.encoder = encoder_net
        self.decoder = decoder_net
        self.threshold_net = threshold_net
        self.threshold_est = Dense(n_features, activation=None)
        self.score_fn = score_fn
        self.beta = beta

    def call(self, x: tf.Tensor) -> tf.Tensor:
        """Forward pass used for teacher-forcing training."""
        # reconstruct input via encoder-decoder
        init_state = self.encoder(x)[1]
        x_recon, z, _ = self.decoder(x, init_state=init_state)

        # compute outlier score
        err_recon = self.score_fn(x, x_recon)

        # estimate outlier threshold from hidden state of decoder
        z = self.threshold_net(z)
        threshold_est = self.threshold_est(z)

        # add threshold estimate loss
        threshold_loss = tf.reduce_mean((err_recon - threshold_est) ** 2)
        self.add_loss(self.beta * threshold_loss)
        return x_recon

    def decode_seq(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Sequence decoding and threshold estimation used for inference."""
        seq_len = np.shape(x)[1]
        n_batch = x.shape[0]

        # use encoder to get state vectors
        init_state = self.encoder(x)[1]

        # generate start of target sequence
        decoder_input = np.reshape(x[:, 0, :], (n_batch, 1, -1))

        # initialize hidden states used to compute outlier thresholds
        z = np.zeros((n_batch, seq_len, init_state[0].numpy().shape[1])).astype(
            np.float32
        )

        # sequential prediction of time series
        decoded_seq = np.zeros_like(x)
        decoded_seq[:, 0, :] = x[:, 0, :]
        i = 1
        while i < seq_len:
            # decode step in sequence
            decoder_output = self.decoder(decoder_input, init_state=init_state)
            decoded_seq[:, i : i + 1, :] = decoder_output[0].numpy()  # noqa E203
            init_state = decoder_output[2]

            # update hidden state decoder used for outlier threshold
            z[:, i : i + 1, :] = decoder_output[1].numpy()  # noqa E203

            # update next decoder input
            decoder_input = np.zeros_like(decoder_input)
            decoder_input[:, :1, :] = decoder_output[0].numpy()

            i += 1

        # compute outlier thresholds
        z = self.threshold_net(z)
        threshold_est = self.threshold_est(z).numpy()

        return decoded_seq, threshold_est


def eucl_cosim_features(x: tf.Tensor, y: tf.Tensor, max_eucl: float = 1e2) -> tf.Tensor:
    """
    Compute features extracted from the reconstructed instance using the
    relative Euclidean distance and cosine similarity between 2 tensors.

    Parameters
    ----------
    x
        Tensor used in feature computation.
    y
        Tensor used in feature computation.
    max_eucl
        Maximum value to clip relative Euclidean distance by.

    Returns
    -------
    Tensor concatenating the relative Euclidean distance and cosine similarity features.
    """
    if len(x.shape) > 2 or len(y.shape) > 2:
        x = Flatten()(x)
        y = Flatten()(y)
    rec_cos = tf.reshape(tf.keras.losses.cosine_similarity(y, x, -1), (-1, 1))
    rec_euc = tf.reshape(relative_euclidean_distance(y, x, -1), (-1, 1))
    # rec_euc could become very large so should be clipped
    rec_euc = tf.clip_by_value(rec_euc, 0, max_eucl)
    return tf.concat([rec_cos, rec_euc], -1)


class AEGMM(tf.keras.Model):
    def __init__(
        self,
        encoder_net: tf.keras.Model,
        decoder_net: tf.keras.Model,
        gmm_density_net: tf.keras.Model,
        n_gmm: int,
        recon_features: Callable = eucl_cosim_features,
        name: str = "aegmm",
    ) -> None:
        """
        Deep Autoencoding Gaussian Mixture Model.

        Parameters
        ----------
        encoder_net
            Layers for the encoder wrapped in a tf.keras.Sequential class.
        decoder_net
            Layers for the decoder wrapped in a tf.keras.Sequential class.
        gmm_density_net
            Layers for the GMM network wrapped in a tf.keras.Sequential class.
        n_gmm
            Number of components in GMM.
        recon_features
            Function to extract features from the reconstructed instance by the decoder.
        name
            Name of the AEGMM model.
        """
        super(AEGMM, self).__init__(name=name)
        self.encoder = encoder_net
        self.decoder = decoder_net
        self.gmm_density = gmm_density_net
        self.n_gmm = n_gmm
        self.recon_features = recon_features

    def call(self, x: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        enc = self.encoder(x)
        x_recon = self.decoder(enc)
        recon_features = self.recon_features(x, x_recon)
        z = tf.concat([enc, recon_features], -1)
        gamma = self.gmm_density(z)
        return x_recon, z, gamma


class VAEGMM(tf.keras.Model):
    def __init__(
        self,
        encoder_net: tf.keras.Model,
        decoder_net: tf.keras.Model,
        gmm_density_net: tf.keras.Model,
        n_gmm: int,
        latent_dim: int,
        recon_features: Callable = eucl_cosim_features,
        beta: float = 1.0,
        name: str = "vaegmm",
    ) -> None:
        """
        Variational Autoencoding Gaussian Mixture Model.

        Parameters
        ----------
        encoder_net
            Layers for the encoder wrapped in a tf.keras.Sequential class.
        decoder_net
            Layers for the decoder wrapped in a tf.keras.Sequential class.
        gmm_density_net
            Layers for the GMM network wrapped in a tf.keras.Sequential class.
        n_gmm
            Number of components in GMM.
        latent_dim
            Dimensionality of the latent space.
        recon_features
            Function to extract features from the reconstructed instance by the decoder.
        beta
            Beta parameter for KL-divergence loss term.
        name
            Name of the VAEGMM model.
        """
        super(VAEGMM, self).__init__(name=name)
        self.encoder = EncoderVAE(encoder_net, latent_dim)
        self.decoder = decoder_net
        self.gmm_density = gmm_density_net
        self.n_gmm = n_gmm
        self.latent_dim = latent_dim
        self.recon_features = recon_features
        self.beta = beta

    def call(self, x: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        enc_mean, enc_log_var, enc = self.encoder(x)
        x_recon = self.decoder(enc)
        recon_features = self.recon_features(x, x_recon)
        z = tf.concat([enc, recon_features], -1)
        gamma = self.gmm_density(z)
        # add KL divergence loss term
        kl_loss = -0.5 * tf.reduce_mean(
            enc_log_var - tf.square(enc_mean) - tf.exp(enc_log_var) + 1
        )
        self.add_loss(self.beta * kl_loss)
        return x_recon, z, gamma



from daml._alibi_detect.od.ae import OutlierAE
from daml._alibi_detect.od.aegmm import OutlierAEGMM
from daml._alibi_detect.od.llr import LLR
from daml._alibi_detect.od.vae import OutlierVAE
from daml._alibi_detect.od.vaegmm import OutlierVAEGMM

__all__ = [
    "OutlierAE",
    "OutlierAEGMM",
    "OutlierVAE",
    "OutlierVAEGMM",
    "LLR",
]



from pathlib import Path
from typing import Any, Callable, Dict, Optional

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
from matplotlib.figure import Figure
from scipy.optimize import minimize
from torch.utils.data import DataLoader, Dataset, Subset


def f_out(n_i: np.ndarray, x: np.ndarray) -> np.ndarray:
    """
    Calculates the line of best fit based on its free parameters

    Parameters
    ----------
    n_i : np.ndarray
        Array of sample sizes
    x : np.ndarray
        Array of inverse power curve coefficients

    Returns
    -------
    np.ndarray
        Data points for the line of best fit
    """
    return x[0] * n_i ** (-x[1]) + x[2]


def calc_params(p_i: np.ndarray, n_i: np.ndarray) -> np.ndarray:
    """
    Retrieves the inverse power curve coefficients for the line of best fit

    Parameters
    ----------
    p_i : np.ndarray
        Array of corresponding losses
    n_i : np.ndarray
        Array of sample sizes

    Returns
    -------
    np.ndarray
        Array of parameters to recreate line of best fit
    """

    def f(x):
        inner = np.square(p_i - x[0] * n_i ** (-x[1]) - x[2])
        return np.sum(inner)

    res = minimize(
        f, np.array([0.5, 0.5, 0.1]), bounds=((0, None), (0, None), (0, None))
    )
    return res.x


def create_data_indices(N: int, M: int) -> np.ndarray:
    """
    Randomly selects integers in range (0, M) an N number of times

    Parameters
    ----------
    N : int
        Size of the first dimension
    M : int
        Size of the second dimension

    Returns
    -------
    np.ndarray
        An NxM array of randomly selected integers in range (0, M)
    """
    return np.random.randint(0, M, size=(N, M))


def reset_parameters(model: nn.Module):
    """
    Re-initializes each layer in the model using
    the layer's defined weight_init function
    """

    @torch.no_grad()
    def weight_reset(m: nn.Module):
        # Check if the current module has reset_parameters
        reset_parameters = getattr(m, "reset_parameters", None)
        if callable(reset_parameters):
            m.reset_parameters()  # type: ignore

    # Applies fn recursively to every submodule see:
    # https://pytorch.org/docs/stable/generated/torch.nn.Module.html
    return model.apply(fn=weight_reset)


class Sufficiency:
    def __init__(
        self,
    ):
        # Train & Eval functions must be set during run
        self._training_func = None
        self._eval_func = None

        self._output_dict = None

    def _train(self, model: nn.Module, dataloader: DataLoader, kwargs: Dict[str, Any]):
        if self._training_func is None:
            raise TypeError("Training function is None. Set function before calling")

        self._training_func(model, dataloader, **kwargs)

    def _eval(self, model: nn.Module, dataloader: DataLoader, kwargs: Dict[str, Any]):
        if self._eval_func is None:
            raise TypeError("Eval function is None. Set function before calling")

        return self._eval_func(model, dataloader, **kwargs)

    def _set_func(self, func: Callable, error_msg="Argument was not a callable"):
        if callable(func):
            return func
        else:
            raise TypeError(error_msg)

    def set_training_func(self, func: Callable):
        self._training_func = self._set_func(func)

    def set_eval_func(self, func: Callable):
        self._eval_func = self._set_func(func)

    def setup(self, length: int, num_models: int = 1, substeps: int = 1):
        if length <= 0:
            raise ValueError("Length cannot be 0")

        # Stores each models' metric output per step
        self._outputs = np.zeros((substeps, num_models))
        # Save the shape for plotting
        self._geomshape = (0.01 * length, length, substeps)
        self._ranges = np.geomspace(*self._geomshape).astype(np.int64)

        # Randomly sample data indices for each model
        self._indices = create_data_indices(num_models, length)

    def run(
        self,
        model: nn.Module,
        train_ds: Dataset,
        test_ds: Dataset,
        batch_size: int = 8,
        train_kwargs: Optional[
            Dict[str, Any]
        ] = None,  # Mutable sequences should not be used as default arg
        eval_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Creates data indices, trains models, and returns plotting data

        Parameters
        ----------
        train : Dataset
            Full training data that will be split for each run
        test : Dataset
            Data that will be used for every run's evaluation
        epochs : int
            Number of training cycles of the dataset, per model
        model_count : int
            Number of models to train and take averages from
        substeps : int
            Total number of dataset partitions that each model will train on

        Returns:
        Dict[str, Any]
        """

        if train_kwargs is None:
            train_kwargs = {}
        if eval_kwargs is None:
            eval_kwargs = {}

        # Bootstrapping
        for j, inds in enumerate(self._indices):
            # Reset the network weights
            model = reset_parameters(model)

            # For each subset of data
            for i, substep in enumerate(self._ranges):
                # We warm start on new data
                b_inds = inds[:substep]

                self._outputs[i, j] = self._run_subset(
                    model,
                    Subset(train_ds, b_inds),
                    test_ds,
                    batch_size,
                    train_kwargs,
                    eval_kwargs,
                )

        n_i = self._ranges
        p_i = 1 - np.mean(self._outputs, axis=1)

        params = calc_params(p_i=p_i, n_i=n_i)

        self._output_dict = {
            "metric": self._outputs,
            "params": params,
            "n_i": n_i,
            "p_i": p_i,
            "geomshape": self._geomshape,
        }

        return self._output_dict

    def _run_subset(
        self,
        model: nn.Module,
        train_data: Dataset,
        eval_data: Dataset,
        batch_size: int,
        train_kwargs,
        eval_kwargs,
    ):
        train_loader = DataLoader(train_data, batch_size=batch_size)
        self._train(model, train_loader, train_kwargs)
        test_loader = DataLoader(eval_data, batch_size=batch_size)
        return self._eval(model, test_loader, eval_kwargs)

    def plot(self, output_dict: Dict[str, Any]):
        """Plotting function for data sufficiency tasks

        Parameters
        ----------
        params : np.ndarray
            The parameters used to calculate the line of best fit
        output_dict : Dict[str, Any]
            Output of sufficiency run

        """
        # Retrieve only relevant values in dictionary
        params = output_dict.get("params", None)
        n_i = output_dict.get("n_i", None)
        p_i = output_dict.get("p_i", None)
        geomshape = output_dict.get("geomshape", None)

        # Ensure all values were retrieved correctly
        if params is None:
            raise KeyError("params not found in output_dict")
        if n_i is None:
            raise KeyError("n_i not found in output_dict")
        if p_i is None:
            raise KeyError("p_i not found in output_dict")
        if geomshape is None:
            raise KeyError("geomshape not found in output_dict")

        # Setup 1 plt figure
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_title("Accuracy vs Number of Images")
        ax.set_xlabel("Number of images")
        ax.set_ylabel("Accuracy")  # TODO: Possibly infer name based on given metric?
        ax.set_ylim(0, 1)

        # Plot model results
        ax.scatter(n_i, 1 - p_i, label="Model Results", s=15, c="black")

        # Plot line of best fit with extrapolation
        extrapolated = np.geomspace(
            geomshape[0], geomshape[1] * 4, geomshape[2]
        ).astype(np.int64)
        ax.plot(
            extrapolated,
            1 - f_out(extrapolated, params),
            linestyle="dashed",
            label="Potential Model Results",
        )

        ax.legend()
        # Save figure
        path = Path("Sufficiency Plot")
        self.save_fig(fig, path)  # type: ignore

    @staticmethod
    def save_fig(fig: Figure, path: Path):
        """
        Saves a `plt.figure` at a given path

        Parameters
        ---------
        fig : plt.figure
            Figure to be saved to png
        path : Path
            Location to save the figure
        Note
        ----
        A directory will be created if it does not exist
        """
        # TODO: Add folder creation, checking, and path handling
        fig.savefig(path)  # type: ignore



"""
This module contains the implementation of Dp Divergence
using the First Nearest Neighbor and Minimum Spanning Tree algorithms
"""
from abc import abstractmethod
from typing import Optional

import numpy as np
import torch
from torch import nn

from daml._internal.metrics.aria.base import _BaseMetric
from daml._internal.metrics.outputs import DivergenceOutput

from .utils import compute_neighbors, minimum_spanning_tree


class _DpDivergence(_BaseMetric):
    """
    For more information about this divergence, its formal definition,
    and its associated estimators
    see https://arxiv.org/abs/1412.6534.
    """

    def __init__(
        self,
        images_a: np.ndarray,
        images_b: np.ndarray,
        encode: bool = False,
        model: Optional[nn.Module] = None,
        fit: Optional[bool] = None,
        epochs: Optional[int] = None,
        device: torch.device = torch.device("cpu"),
    ) -> None:
        """Constructor method"""
        super().__init__(
            images_a,
            encode,
            model=model,
            fit=fit,
            epochs=epochs,
            device=device,
        )
        self.images_a = images_a
        self.images_b = images_b

    @abstractmethod
    def calculate_errors(self, data: np.ndarray, labels: np.ndarray) -> int:
        """Abstract method for the implementation of divergence calculation"""

    def _encode_and_vstack(self) -> np.ndarray:
        emb_a = self._encode(self.images_a)
        emb_b = self._encode(self.images_b)
        if self.encode:
            emb_a = emb_a.flatten()
            emb_b = emb_b.flatten()
        return np.vstack((emb_a, emb_b))

    def _evaluate(self) -> DivergenceOutput:
        """
        Calculates the divergence and any errors between the datasets

        Returns
        -------
        DivergenceOutput
            Dataclass containing the dp divergence and errors during calculation
        """
        N = self.images_a.shape[0]
        M = self.images_b.shape[0]

        images = self._encode_and_vstack()
        labels = np.vstack([np.zeros([N, 1]), np.ones([M, 1])])

        errors = self.calculate_errors(images, labels)
        dp = 1 - ((M + N) / (2 * M * N)) * errors
        return DivergenceOutput(dpdivergence=dp, error=errors)


class DpDivergenceMST(_DpDivergence):
    """
    A minimum spanning tree implementation of dp divergence

    Warning
    -------
        MST is very slow in this implementation, this is unlike matlab where
        they have comparable speeds
        Overall, MST takes ~25x LONGER!!
        Source of slowdown:
        conversion to and from CSR format adds ~10% of the time diff between
        1nn and scipy mst function the remaining 90%
    """

    # TODO
    # - validate the input algorithm
    # - improve speed for MST, requires a fast mst implementation
    # mst is at least 10x slower than knn approach

    def calculate_errors(self, data: np.ndarray, labels: np.ndarray) -> int:
        """
        Returns the divergence between two datasets using a minimum spanning tree

        Parameters
        ----------
        data : np.ndarray
            Array containing images from two datasets
        labels : np.ndarray
            Array showing which dataset each image belonged to

        Returns
        -------
        int
            Number of edges connecting the two datasets

        Note
        ----
        We add a small constant to the distance matrix to ensure scipy interprets
        the input graph as fully-connected.
        """
        mst = minimum_spanning_tree(data).toarray()
        edgelist = np.transpose(np.nonzero(mst))
        errors = np.sum(labels[edgelist[:, 0]] != labels[edgelist[:, 1]])
        return errors


class DpDivergenceFNN(_DpDivergence):
    """
    A first nearest neighbor implementation of dp divergence
    """

    def calculate_errors(self, data: np.ndarray, labels: np.ndarray) -> int:
        """
        Returns the divergence between two datasets using first nearest neighbor

        Parameters
        ----------
        data : np.ndarray
            Array containing images from two datasets
        labels : np.ndarray
            Array showing which dataset each image belonged to

        Returns
        -------
        int
            Number of edges connecting the two datasets
        """
        nn_indices = compute_neighbors(data, data)
        errors = np.sum(np.abs(labels[nn_indices] - labels))
        return errors



"""
Defining types compatible with different Python versions and defining custom types.
"""
from typing import Optional, Type, Union

from sklearn.base import (  # import here (instead of later) since sklearn currently a core dep
    BaseEstimator,
)

# Literal for typing
from typing_extensions import Literal, TypeAlias

from daml._alibi_detect.utils.frameworks import has_pytorch, has_tensorflow

# Optional dep dependent tuples of types, for isinstance checks and pydantic
supported_models_tf: tuple = ()
supported_models_torch: tuple = ()
supported_optimizers_tf: tuple = ()
supported_optimizers_torch: tuple = ()
if has_tensorflow:
    import tensorflow as tf

    supported_models_tf = (tf.keras.Model,)
    if hasattr(tf.keras.optimizers, "legacy"):
        supported_optimizers_tf = (
            tf.keras.optimizers.Optimizer,
            tf.keras.optimizers.legacy.Optimizer,
            type,
        )
    else:
        supported_optimizers_tf = (tf.keras.optimizers.Optimizer, type)
if has_pytorch:
    import torch

    supported_models_torch = (torch.nn.Module,)
    supported_optimizers_torch = (type,)  # Note type not object!
supported_models_sklearn = (BaseEstimator,)
supported_models_all = (
    supported_models_tf + supported_models_torch + supported_models_sklearn
)
supported_optimizers_all = supported_optimizers_tf + supported_optimizers_torch

# type aliases, for use with mypy (must be FwdRef's if involving opt. deps.)
OptimizerTF: TypeAlias = Union[
    "tf.keras.optimizers.Optimizer",
    "tf.keras.optimizers.legacy.Optimizer",
    Type["tf.keras.optimizers.Optimizer"],
    Type["tf.keras.optimizers.legacy.Optimizer"],
]

TorchDeviceType: TypeAlias = Optional[
    Union[Literal["cuda", "gpu", "cpu"], "torch.device"]
]



from daml._internal.metrics.alibi_detect.ae import AlibiAE as OD_AE
from daml._internal.metrics.alibi_detect.aegmm import AlibiAEGMM as OD_AEGMM
from daml._internal.metrics.alibi_detect.llr import AlibiLLR as OD_LLR
from daml._internal.metrics.alibi_detect.vae import AlibiVAE as OD_VAE
from daml._internal.metrics.alibi_detect.vaegmm import AlibiVAEGMM as OD_VAEGMM
from daml._internal.metrics.outputs import OutlierDetectorOutput
from daml._internal.metrics.types import Threshold, ThresholdType

__all__ = [
    "OD_AE",
    "OD_AEGMM",
    "OD_LLR",
    "OD_VAE",
    "OD_VAEGMM",
    "OutlierDetectorOutput",
    "Threshold",
    "ThresholdType",
]



import maite.protocols as pr
import numpy as np

from daml._internal.datasets.datasets import DamlDataset


class JaticClassificationDatasetWrapper(DamlDataset):
    "Reformats Jatic data into the DAML format"

    def __init__(self, dataset: pr.VisionDataset):
        images, labels = self._split_data(dataset)
        super().__init__(images, labels)

    def _split_data(self, dataset: pr.VisionDataset):
        images = []
        labels = []
        for idx in range(len(dataset)):
            images.append(np.array(dataset[idx]["image"]))
            labels.append(np.array(dataset[idx]["label"]))

        images = np.array(images)
        labels = np.array(labels)

        return images, labels


class JaticObjectDetectionWrapper(DamlDataset):
    "Reformats Jatic data into the DAML format"

    def __init__(self, dataset: pr.ObjectDetectionDataset):
        images, labels, boxes = self._split_data(dataset)
        super().__init__(images=images, labels=labels, boxes=boxes)

    def _split_data(self, dataset: pr.ObjectDetectionDataset):
        images = []
        labels = []
        boxes = []
        for idx in range(len(dataset)):
            objects: pr.HasDataBoxesLabels = dataset[idx]["objects"]  # type: ignore

            images.append(np.array(dataset[idx]["image"]))
            labels.append(np.array(objects["labels"]))
            boxes.append(np.array(objects["boxes"]))

        images = np.array(images)
        labels = np.array(labels)
        boxes = np.array(boxes)

        return images, labels, boxes



import logging
from typing import Callable, Dict, Tuple

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

from daml._alibi_detect.base import (
    BaseDetector,
    FitMixin,
    ThresholdMixin,
    outlier_prediction_dict,
)
from daml._alibi_detect.models.tensorflow.autoencoder import VAEGMM, eucl_cosim_features
from daml._alibi_detect.models.tensorflow.gmm import gmm_energy, gmm_params
from daml._alibi_detect.models.tensorflow.losses import loss_vaegmm
from daml._alibi_detect.models.tensorflow.trainer import trainer
from daml._alibi_detect.utils._types import OptimizerTF
from daml._alibi_detect.utils.tensorflow.prediction import predict_batch

logger = logging.getLogger(__name__)


class OutlierVAEGMM(BaseDetector, FitMixin, ThresholdMixin):
    def __init__(
        self,
        threshold: float = None,
        vaegmm: tf.keras.Model = None,
        encoder_net: tf.keras.Model = None,
        decoder_net: tf.keras.Model = None,
        gmm_density_net: tf.keras.Model = None,
        n_gmm: int = None,
        latent_dim: int = None,
        samples: int = 10,
        beta: float = 1.0,
        recon_features: Callable = eucl_cosim_features,
        data_type: str = None,
    ) -> None:
        """
        VAEGMM-based outlier detector.

        Parameters
        ----------
        threshold
            Threshold used for outlier score to determine outliers.
        vaegmm
            A trained tf.keras model if available.
        encoder_net
            Layers for the encoder wrapped in a tf.keras.Sequential class if no 'vaegmm' is specified.
        decoder_net
            Layers for the decoder wrapped in a tf.keras.Sequential class if no 'vaegmm' is specified.
        gmm_density_net
            Layers for the GMM network wrapped in a tf.keras.Sequential class.
        n_gmm
            Number of components in GMM.
        latent_dim
            Dimensionality of the latent space.
        samples
            Number of samples sampled to evaluate each instance.
        beta
            Beta parameter for KL-divergence loss term.
        recon_features
            Function to extract features from the reconstructed instance by the decoder.
        data_type
            Optionally specify the data type (tabular, image or time-series). Added to metadata.
        """
        super().__init__()

        if threshold is None:
            logger.warning(
                "No threshold level set. Need to infer threshold using `infer_threshold`."
            )

        self.threshold = threshold
        self.samples = samples

        # check if model can be loaded, otherwise initialize VAEGMM model
        if isinstance(vaegmm, tf.keras.Model):
            self.vaegmm = vaegmm
        elif (
            isinstance(encoder_net, tf.keras.Sequential)
            and isinstance(decoder_net, tf.keras.Sequential)
            and isinstance(gmm_density_net, tf.keras.Sequential)
        ):
            self.vaegmm = VAEGMM(
                encoder_net,
                decoder_net,
                gmm_density_net,
                n_gmm,
                latent_dim,
                recon_features=recon_features,
                beta=beta,
            )
        else:
            raise TypeError(
                "No valid format detected for `vaegmm` (tf.keras.Model) "
                "or `encoder_net`, `decoder_net` and `gmm_density_net` (tf.keras.Sequential)."
            )

        # set metadata
        self.meta["detector_type"] = "outlier"
        self.meta["data_type"] = data_type
        self.meta["online"] = False

        self.phi, self.mu, self.cov, self.L, self.log_det_cov = (
            None,
            None,
            None,
            None,
            None,
        )

    def fit(
        self,
        X: np.ndarray,
        loss_fn: tf.keras.losses = loss_vaegmm,
        w_recon: float = 1e-7,
        w_energy: float = 0.1,
        w_cov_diag: float = 0.005,
        optimizer: OptimizerTF = tf.keras.optimizers.Adam,
        cov_elbo: dict = dict(sim=0.05),
        epochs: int = 20,
        batch_size: int = 64,
        verbose: bool = True,
        log_metric: Tuple[str, "tf.keras.metrics"] = None,
        callbacks: tf.keras.callbacks = None,
    ) -> None:
        """
        Train VAEGMM model.

        Parameters
        ----------
        X
            Training batch.
        loss_fn
            Loss function used for training.
        w_recon
            Weight on elbo loss term if default `loss_vaegmm`.
        w_energy
            Weight on sample energy loss term if default `loss_vaegmm` loss fn is used.
        w_cov_diag
            Weight on covariance regularizing loss term if default `loss_vaegmm` loss fn is used.
        optimizer
            Optimizer used for training.
        cov_elbo
            Dictionary with covariance matrix options in case the elbo loss function is used.
            Either use the full covariance matrix inferred from X (dict(cov_full=None)),
            only the variance (dict(cov_diag=None)) or a float representing the same standard deviation
            for each feature (e.g. dict(sim=.05)).
        epochs
            Number of training epochs.
        batch_size
            Batch size used for training.
        verbose
            Whether to print training progress.
        log_metric
            Additional metrics whose progress will be displayed if verbose equals True.
        callbacks
            Callbacks used during training.
        """
        # train arguments
        args = [self.vaegmm, loss_fn, X]
        optimizer = optimizer() if isinstance(optimizer, type) else optimizer
        kwargs = {
            "optimizer": optimizer,
            "epochs": epochs,
            "batch_size": batch_size,
            "verbose": verbose,
            "log_metric": log_metric,
            "callbacks": callbacks,
            "loss_fn_kwargs": {
                "w_recon": w_recon,
                "w_energy": w_energy,
                "w_cov_diag": w_cov_diag,
            },
        }

        # initialize covariance matrix if default vaegmm loss fn is used
        use_elbo = loss_fn.__name__ == "loss_vaegmm"
        cov_elbo_type, cov = [*cov_elbo][0], [*cov_elbo.values()][0]
        if use_elbo and cov_elbo_type in ["cov_full", "cov_diag"]:
            cov = tfp.stats.covariance(X.reshape(X.shape[0], -1))
            if (
                cov_elbo_type == "cov_diag"
            ):  # infer standard deviation from covariance matrix
                cov = tf.math.sqrt(tf.linalg.diag_part(cov))
        if use_elbo:
            kwargs["loss_fn_kwargs"][cov_elbo_type] = tf.dtypes.cast(cov, tf.float32)

        # train
        trainer(*args, **kwargs)

        # set GMM parameters
        x_recon, z, gamma = self.vaegmm(X)
        self.phi, self.mu, self.cov, self.L, self.log_det_cov = gmm_params(z, gamma)

    def infer_threshold(
        self, X: np.ndarray, threshold_perc: float = 95.0, batch_size: int = int(1e10)
    ) -> None:
        """
        Update threshold by a value inferred from the percentage of instances considered to be
        outliers in a sample of the dataset.

        Parameters
        ----------
        X
            Batch of instances.
        threshold_perc
            Percentage of X considered to be normal based on the outlier score.
        batch_size
            Batch size used when making predictions with the VAEGMM.
        """
        # compute outlier scores
        iscore = self.score(X, batch_size=batch_size)

        # update threshold
        self.threshold = np.percentile(iscore, threshold_perc)

    def score(self, X: np.ndarray, batch_size: int = int(1e10)) -> np.ndarray:
        """
        Compute outlier scores.

        Parameters
        ----------
        X
            Batch of instances to analyze.
        batch_size
            Batch size used when making predictions with the VAEGMM.

        Returns
        -------
        Array with outlier scores for each instance in the batch.
        """
        # draw samples from latent space
        X_samples = np.repeat(X, self.samples, axis=0)
        _, z, _ = predict_batch(X_samples, self.vaegmm, batch_size=batch_size)

        # compute average energy for samples
        energy, _ = gmm_energy(
            z, self.phi, self.mu, self.cov, self.L, self.log_det_cov, return_mean=False
        )
        energy_samples = energy.numpy().reshape((-1, self.samples))
        iscore = np.mean(energy_samples, axis=-1)
        return iscore

    def predict(
        self,
        X: np.ndarray,
        batch_size: int = int(1e10),
        return_instance_score: bool = True,
    ) -> Dict[Dict[str, str], Dict[np.ndarray, np.ndarray]]:
        """
        Compute outlier scores and transform into outlier predictions.

        Parameters
        ----------
        X
            Batch of instances.
        batch_size
            Batch size used when making predictions with the VAEGMM.
        return_instance_score
            Whether to return instance level outlier scores.

        Returns
        -------
        Dictionary containing ``'meta'`` and ``'data'`` dictionaries.
            - ``'meta'`` has the model's metadata.
            - ``'data'`` contains the outlier predictions and instance level outlier scores.
        """
        # compute outlier scores
        iscore = self.score(X, batch_size=batch_size)

        # values above threshold are outliers
        outlier_pred = (iscore > self.threshold).astype(int)

        # populate output dict
        od = outlier_prediction_dict()
        od["meta"] = self.meta
        od["data"]["is_outlier"] = outlier_pred
        if return_instance_score:
            od["data"]["instance_score"] = iscore
        return od



import numpy as np
import tensorflow as tf


def mutate_categorical(
    X: np.ndarray, rate: float = None, seed: int = 0, feature_range: tuple = (0, 255)
) -> tf.Tensor:
    """
    Randomly change integer feature values to values within a set range
    with a specified permutation rate.

    Parameters
    ----------
    X
        Batch of data to be perturbed.
    rate
        Permutation rate (between 0 and 1).
    seed
        Random seed.
    feature_range
        Min and max range for perturbed features.

    Returns
    -------
    Array with perturbed data.
    """
    frange = (feature_range[0] + 1, feature_range[1] + 1)
    shape = X.shape
    n_samples = np.prod(shape)
    mask = tf.random.categorical(
        tf.math.log([[1.0 - rate, rate]]), n_samples, seed=seed, dtype=tf.int32
    )
    mask = tf.reshape(mask, shape)
    possible_mutations = tf.random.uniform(
        shape, minval=frange[0], maxval=frange[1], dtype=tf.int32, seed=seed + 1
    )
    X = tf.math.floormod(tf.cast(X, tf.int32) + mask * possible_mutations, frange[1])
    return tf.cast(X, tf.float32)



from daml._internal.metrics.aria.divergence import DpDivergenceFNN as HP_FNN
from daml._internal.metrics.aria.divergence import DpDivergenceMST as HP_MST
from daml._internal.metrics.outputs import DivergenceOutput

__all__ = ["HP_FNN", "HP_MST", "DivergenceOutput"]



from daml._internal.metrics.aria.ber import MultiClassBerFNN as BER_FNN
from daml._internal.metrics.aria.ber import MultiClassBerMST as BER_MST
from daml._internal.metrics.outputs import BEROutput

__all__ = ["BER_FNN", "BER_MST", "BEROutput"]



from typing import Optional

from daml import _alibi_detect
from daml._internal.metrics.alibi_detect.base import (
    AlibiDetectOutlierType,
    _AlibiDetectMetric,
)


class AlibiVAE(_AlibiDetectMetric):
    """
    Variational Autoencoder-based outlier detector,
    using `alibi-detect vae. <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/vae.html>`_

    The model used by this class is :py:class:`daml.models.VAE`
    """  # noqa E501

    def __init__(self):
        super().__init__(
            alibi_detect_class=_alibi_detect.od.OutlierVAE,
            model_param_name="vae",
            flatten_dataset=False,
            dataset_type=None,
        )

    def set_prediction_args(
        self,
        outlier_type: Optional[AlibiDetectOutlierType] = None,
        outlier_perc: Optional[float] = None,
        return_feature_score: Optional[bool] = None,
        return_instance_score: Optional[bool] = None,
        batch_size: Optional[int] = None,
    ) -> None:
        """
        Sets additional arguments to be used during prediction.

        Note
        ----
        Visit `alibi-detect vae <https://docs.seldon.io/projects/alibi-detect/en/latest/od/methods/vae.html#Detect>`_ for additional information on prediction parameters.
        """  # noqa E501
        self._update_kwargs_with_locals(self._predict_kwargs, **locals())

    @property
    def _default_predict_kwargs(self) -> dict:
        return {
            "outlier_type": AlibiDetectOutlierType.INSTANCE,
            "outlier_perc": 75,
            "return_feature_score": True,
            "return_instance_score": True,
            "batch_size": 64,
        }
